from mmcif_gen.investigation_engine import InvestigationEngine
from mmcif_gen.investigation_io import SqliteReader
from typing import List
import sys
import logging
import argparse

logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)

class InvestigationMaxIV(InvestigationEngine):
        
    def __init__(self, sqlite_path: str, investigation_id: str, output_path: str, maxiv_investigation_json: str="./operations/maxiv/maxiv_investigation.json") -> None:
        logging.info("Instantiating MaxIV Investigation subclass")
        self.reader = SqliteReader(sqlite_path)
        self.operation_file_json = maxiv_investigation_json
        super().__init__(investigation_id, output_path)

    def pre_run(self) -> None:
        logging.info("Pre-running")
        self.create_denormalized_table()
        self.create_entity_table()
        self.add_descript_categories()
        super().pre_run()

    def create_denormalized_table(self) -> None:
        logging.info("Creating denormalized table")
        query = "DROP TABLE IF EXISTS denormalized_data"
        self.reader.sql_execute(query)
        # Remove rows where S.data_collection_type is 'dummy', as no crystal is in the beam.
        query ='''CREATE TABLE denormalized_data AS 
        SELECT 
            S.dataset_id, S.session, S.data_collection_date, S.data_collection_outcome,
            A.mounted_crystal_id, A.cryo_chem_comp_code,
            B.crystal_plate_id, B.marked_crystal_id, B.crystal_screen_condition_id,
            C.soak_plate_id,
            E.crystal_plate_id, E.protein_batch_id,
            F.protein_batch_sequence, F.protein_batch_comp_id, F.protein_batch_uniprot_id, F.protein_acronym,
            G.solvent, G.compound_code, G.compound_batch_code, G.library_name,
            H.smiles, H.inchi, H.chemical_name, H.formula, H.formula_weight, H.cas,
            I.protein_name, I.proposal_number,
            J.crystal_screen_id, J.crystal_screen_chem_comp_ids
        FROM xray_dataset_table S
        INNER JOIN mounted_crystals_table A on S.mounted_crystal_id = A.mounted_crystal_id
            INNER JOIN marked_crystals_table B on A.marked_crystal_id = B.marked_crystal_id
                LEFT JOIN soaked_crystals_table C on B.marked_crystal_id = C.marked_crystal_id 
                    INNER JOIN soak_plate_table D on C.soak_plate_id = D.soak_plate_id
                        INNER JOIN compound_batch_table G on G.compound_batch_code = D.compound_batch_code
                            INNER JOIN compound_table H on H.compound_code = G.compound_code
                INNER JOIN crystal_plate_table E on B.crystal_plate_id = E.crystal_plate_id
                    INNER JOIN protein_batch_table F on E.protein_batch_id = F.protein_batch_id
                    INNER JOIN project_table I on I.protein_acronym = F.protein_acronym
                INNER JOIN crystal_screen_condition_table J on B.crystal_screen_condition_id = J.crystal_screen_condition_id
                WHERE  G.library_name  NOT IN ('solvent test', 'SU','solvent_test','solvent_only','SU_FU_2023-12-07','solvent_tolerance_231211')
        '''
        self.reader.sql_execute(query)

    def create_entity_table(self) -> None:
        logging.info("Creating entity table")
        with self.reader.sqlite_db_connection() as cursor:
            query = "DROP TABLE IF EXISTS entities"
            cursor.execute(query)
            query = '''
                    CREATE TABLE entities (
                        entity_id TEXT,
                        type TEXT,
                        sequence TEXT,
                        chem_comp_id TEXT,
                        name TEXT,
                        formula TEXT,
                        formula_weight TEXT,
                        inchi TEXT,
                        uniprot_id TEXT,
                        sample_id INT,
                        campaign_id TEXT,
                        series_id TEXT
                )
            '''
            cursor.execute(query)

            chem_comp_query = '''WITH RECURSIVE split1(crystal_screen_id, code, str) 
                    AS (
                        SELECT distinct(crystal_screen_id), '', crystal_screen_chem_comp_ids ||' ' as code FROM denormalized_data
                        UNION ALL SELECT
                        crystal_screen_id,
                        substr(str, 0, instr(str, ' ')),
                        substr(str, instr(str, ' ')+1)
                        FROM split1 WHERE str!=''
                    ),
                split2(protein_batch_id, code, str) 
                    AS (
                        SELECT distinct(protein_batch_id), '', protein_batch_comp_id ||' ' as code FROM denormalized_data
                        UNION ALL SELECT
                        protein_batch_id,
                        substr(str, 0, instr(str, ' ')),
                        substr(str, instr(str, ' ')+1)
                        FROM split2 WHERE str!=''
                    ),
                split3(mounted_crystal_id, code, str) 
                    AS (
                        SELECT distinct(mounted_crystal_id), '', cryo_chem_comp_code ||' ' as code FROM denormalized_data
                        UNION ALL SELECT
                        mounted_crystal_id,
                        substr(str, 0, instr(str, ' ')),
                        substr(str, instr(str, ' ')+1)
                        FROM split3 WHERE str!=''
                    ), 
                combined as (
                    Select code from split1
                    UNION
                    Select code from split2
                    UNION
                    Select code from split3
                    )
                SELECT DISTINCT(code), c.name, c.formula, c.formula_weight, c.inchi from combined LEFT JOIN wwpdb_chem_comp_table c on c.chem_comp_code = combined.code where combined.code != '' 
            '''
            results = self.reader.sql_execute(chem_comp_query)

            for index, result in enumerate(results):
                insert_query = """
                        INSERT INTO entities
                        (entity_id, type, chem_comp_id,name, formula, formula_weight, inchi )
                        VALUES (?, ?, ?, ?, ?, ?, ?)
                    """
                cursor.execute(
                        insert_query,
                        ( str(index+1),
                            "non-polymer",
                            result[0],
                            result[1],
                            result[2],
                            result[3],
                            result[4]
                        ),
                        )
            poly_query = '''
                        WITH RECURSIVE split(protein_batch_uniprot_id, seq, str) AS 
                        (SELECT distinct(protein_batch_uniprot_id), '', protein_batch_sequence ||'\n' as code 
                        FROM protein_batch_table 
                        UNION ALL 
                        SELECT protein_batch_uniprot_id, 
                        substr(str, 0, instr(str, '\n')), 
                        substr(str, instr(str, '\n')+1) 
                        FROM split WHERE str!='' ) 
                        SELECT seq, protein_batch_uniprot_id FROM split WHERE length(seq)> 7'''
            
            results = self.reader.sql_execute(poly_query)
            for index, result in enumerate(results):
                insert_query = """
                        INSERT INTO entities
                        (entity_id, sequence, type, uniprot_id )
                        VALUES (?, ?, ?, ?)
                    """
                cursor.execute(insert_query,
                                ( str(index+1),
                                    result[0],
                                    "polymer",
                                    result[1]
                                ),)

            

    def add_descript_categories(self):
        logging.info("Adding Descript categories")
        query = "DROP TABLE IF EXISTS descript"
        self.reader.sql_execute(query)
        query_creating_descript = '''
                                    CREATE TABLE descript (
                                    entity_id int,
                                    fragment_component_id int,
                                    fragment_component_mix int,
                                    dataset_id int,
                                    chem_comp text,
                                    sequence text,
                                    fragment_inchi text,
                                    poly_descript int,
                                    nonpoly_descript int,
                                    sample_id int)
                                    '''
        self.reader.sql_execute(query_creating_descript)

        query_nonpoly_descript = '''
                                    SELECT  protein_batch_comp_id, crystal_screen_chem_comp_ids, cryo_chem_comp_code from denormalized_data 
                                    group by protein_batch_comp_id, crystal_screen_chem_comp_ids, cryo_chem_comp_code
                                    '''
        nonpoly_result = self.reader.sql_execute(query_nonpoly_descript)

        query_poly_descript = "SELECT  entity_id, sequence FROM entities WHERE type='polymer'"
        poly_result = self.reader.sql_execute(query_poly_descript)

        query_poly_mapping = "SELECT distinct(protein_batch_sequence) FROM protein_batch_table"
        poly_mapping_result = self.reader.sql_execute(query_poly_mapping)

        chem_comp_entity_id_query = "SELECT entity_id, chem_comp_id FROM entities WHERE type='non-polymer' "
        nonpoly_mapping_result = self.reader.sql_execute(chem_comp_entity_id_query)

        query_sample = '''
                            SELECT  protein_batch_sequence, 
                            COALESCE(protein_batch_comp_id, '') || ' ' || COALESCE(crystal_screen_chem_comp_ids,'') || ' ' || COALESCE(cryo_chem_comp_code,'') 
                            FROM denormalized_data 
                            group by protein_batch_comp_id, crystal_screen_chem_comp_ids, cryo_chem_comp_code, protein_batch_sequence
                        '''
        sample_result = self.reader.sql_execute(query_sample)

        query_fragment = '''
                SELECT inchi, dataset_id from denormalized_data where inchi is not null and dataset_id is not NULL ORDER BY dataset_id
            '''
        fragment_result = self.reader.sql_execute(query_fragment)

        mapping = {}
        for row in nonpoly_mapping_result:
            mapping[row[1]] = row[0]
        
        with self.reader.sqlite_db_connection() as cursor:
            for sample_index, entities in enumerate(sample_result):
                non_poly_sample_entities = entities[1].split()
                for descript, chem_comp_codes in enumerate(nonpoly_result):
                    nonpoly_descript = []
                    for group in chem_comp_codes:
                        if group:
                            nonpoly_descript += group.split()
                    for chem_comp_id in nonpoly_descript:
                        sample_id = None
                        if chem_comp_id in non_poly_sample_entities:
                            sample_id = sample_index+1
                        insert_query = '''
                            INSERT INTO descript
                            (entity_id, chem_comp, nonpoly_descript, sample_id )
                            VALUES (?, ?, ?, ?)'''
                        cursor.execute(insert_query, (mapping[chem_comp_id],chem_comp_id, descript+1, sample_id))
                for seq in poly_result:
                    for  descript, row in enumerate(poly_mapping_result):
                        if seq[1] in row[0]:
                            sample_id = None
                            if seq[1] in entities[0]:
                                sample_id = sample_index+1
                            insert_query = '''
                                    INSERT INTO descript
                                    (entity_id, sequence, poly_descript, sample_id )
                                    VALUES (?, ?, ?, ?)'''
                            cursor.execute(insert_query, (seq[0],seq[1], descript+1, sample_id))

            previous_dataset_id = ""
            frag_component_mix = 0
            fragment_component_id_mapping = {}
            next_component_id = 1
            fragmix_data = []
            frag_data = []
            mix_for_index = []
            for index, fragment_data in enumerate(fragment_result):
                if fragment_data[1] != previous_dataset_id:
                    frag_component_mix += 1

                if fragment_data[0]  not in fragment_component_id_mapping:
                    fragment_component_id_mapping[fragment_data[0]] = next_component_id
                    next_component_id += 1

                fragment_component_id = fragment_component_id_mapping[fragment_data[0]]
                
                fragmix_data.append((frag_component_mix, fragment_component_id))
                mix_for_index.append(frag_component_mix)
                frag_data.append((fragment_data[0],fragment_data[1]))
                
                previous_dataset_id = fragment_data[1]

            # Finding all mixes:

            mix_to_fragments = {}
            for mix, fragment in fragmix_data:
                if mix not in mix_to_fragments:
                    mix_to_fragments[mix] = set()
                mix_to_fragments[mix].add(fragment)

            # Finding Unique mixes
            unique_fragments = {}
            mix_id = 0 
            for mix, fragments in mix_to_fragments.items():
                fragment_tuple = tuple(sorted(fragments))  # Sort and convert to tuple for hashability
                if fragment_tuple not in unique_fragments:
                    index_in_frag_data = mix_for_index.index(mix)
                    unique_fragments[fragment_tuple] = mix
                    mix_id += 1
                    for frag in fragment_tuple:
                        insert_query = '''
                                INSERT INTO descript
                                (fragment_component_id, fragment_component_mix, fragment_inchi, dataset_id)
                                VALUES (?, ?, ?, ?)'''
                        cursor.execute(insert_query, (frag, mix_id, frag_data[index_in_frag_data][0],frag_data[index_in_frag_data][1]))




def run(sqlite_path : str, investigation_id: str, output_path: str) -> None:
    im = InvestigationMaxIV(sqlite_path, investigation_id, output_path)
    im.pre_run()
    im.run()
    
def maxiv_subparser(subparsers, parent_parser):
    parser_maxiv = subparsers.add_parser("maxiv",help="Parameter requirements for investigation files from MAX IV data",parents=[parent_parser])
    parser_maxiv.add_argument(
        "-s",
        "--sqlite",
        help="Path to the Sqlite DB for the given investigation",
    )

def run_investigation_maxiv(args):
    if not args.sqlite:
        logging.error("Max IV facility requires path to --sqlite file")
        return 1
    run(args.sqlite, args.id, args.output_folder,args.json)


    


from mmcif_gen.facilities.pdbe import pdbe_subparser, run_investigation_pdbe
from mmcif_gen.facilities.maxiv import maxiv_subparser, run_investigation_maxiv
from mmcif_gen.facilities.dls import dls_subparser, run_investigation_dls
from mmcif_gen.facilities.xchem import xchem_subparser, run_investigation_xchem
import argparse
import json
import logging
from logging.handlers import RotatingFileHandler

import os
import pathlib
import requests
import sys
from typing import Dict, List, Optional

file_handler = RotatingFileHandler('mmcifgen.log', maxBytes=100000, backupCount=3)
file_handler.setFormatter(logging.Formatter('%(asctime)s - %(levelname)s - %(message)s'))
file_handler.setLevel(logging.DEBUG)

logging.getLogger().addHandler(file_handler)

FACILITIES_URL = "https://raw.githubusercontent.com/PDBeurope/Investigations/main/mmcif_gen/operations/fetched_list.json"

class CLIManager:
    def __init__(self):
        self.facilities_data = None
        
    def fetch_facilities_data(self) -> Dict:
        """Fetch facilities data from GitHub."""
        if self.facilities_data is None:
            try:
                response = requests.get(FACILITIES_URL)
                response.raise_for_status()
                self.facilities_data = response.json()
            except requests.RequestException as e:
                print(f"Error fetching facilities data: {e}", file=sys.stderr)
                sys.exit(1)
        return self.facilities_data

    def get_available_facilities(self) -> List[str]:
        """Get list of available facilities."""
        return ['pdbe', 'maxiv', 'dls', 'xchem']
        # return list(self.fetch_facilities_data().keys())

    def get_facility_jsons(self, facility: str) -> List[str]:
        """Get available JSON files for a facility."""
        data = self.fetch_facilities_data()
        return data.get(facility, [])

    def fetch_facility_json(self, json_path: str, output_dir: str = ".") -> str:
        """Fetch a specific facility JSON file."""
        base_url = "https://raw.githubusercontent.com/PDBeurope/Investigations/main/mmcif_gen/"
        full_url = base_url + json_path
        
        os.makedirs(output_dir, exist_ok=True)
        filename = os.path.basename(json_path)
        output_path = os.path.join(output_dir, filename)
        
        print(f"Fetching {filename}...")
        try:
            response = requests.get(full_url)
            response.raise_for_status()
            
            with open(output_path, 'w') as f:
                json.dump(response.json(), f, indent=2)
            
            print(f"Successfully saved to {output_path}")
            return output_path
        except requests.RequestException as e:
            print(f"Error fetching JSON: {e}", file=sys.stderr)
            sys.exit(1)

    def find_local_json(self, facility: str) -> Optional[str]:
        """Find facility JSON in current directory."""
        possible_files = [
            f"{facility}_metadata.json",
            f"{facility}_metadata_hardcoded.json",
            f"{facility}_operations.json",
            f"{facility}_investigation.json"
        ]
        
        for file in possible_files:
            if os.path.exists(file):
                return file
        return None

def setup_parsers():
    parser = argparse.ArgumentParser(
        prog="mmcif-gen",
        description="Generate mmCIF files from various facility data sources"
    )
    subparsers = parser.add_subparsers(dest="command", help="Available commands")

    # Common arguments for make-mmcif command
    make_mmcif_parent = argparse.ArgumentParser(add_help=False)
    make_mmcif_parent.add_argument(
        "--json",
        help="Path to transformation JSON file"
    )
    make_mmcif_parent.add_argument(
        "--output-folder",
        help="Output folder for mmCIF files",
        default="./out"
    )
    make_mmcif_parent.add_argument(
        "--id",
        help="File identifier",
        default="I_1234"
    )

    # fetch-facility-json command
    fetch_parser = subparsers.add_parser(
        "fetch-facility-json",
        help="Fetch facility JSON configuration"
    )
    fetch_parser.add_argument(
        "json_name",
        help="Name of the JSON to fetch (e.g. dls-metadata)"
    )
    fetch_parser.add_argument(
        "-o", "--output-dir",
        help="Output directory",
        default="."
    )

    # make-mmcif command
    make_parser = subparsers.add_parser(
        "make-mmcif",
        help="Generate mmCIF file",
        parents=[make_mmcif_parent]
    )

    make_mmcif_subparser = make_parser.add_subparsers(
        dest="facility",
        help="Specifies facility for which mmcif files will be used for",
        required=True
    )
   
    # Create facility-specific parent parser
    facility_parent = argparse.ArgumentParser(add_help=False)
    
    # Add facility subparsers
    pdbe_subparser(make_mmcif_subparser, facility_parent)
    maxiv_subparser(make_mmcif_subparser, facility_parent)
    dls_subparser(make_mmcif_subparser, facility_parent)
    xchem_subparser(make_mmcif_subparser, facility_parent)

    return parser

def handle_command_error(parser: argparse.ArgumentParser, message: str):
    """Handle command errors with user-friendly messages."""
    print(f"Error: {message}", file=sys.stderr)
    parser.print_help()
    sys.exit(1)

def main():
    parser = setup_parsers()
    
    # Convert underscores to hyphens in command
    if len(sys.argv) > 1:
        sys.argv[1] = sys.argv[1].replace('_', '-')
    
    args = parser.parse_args()
    cli_manager = CLIManager()

    if not args.command:
        parser.print_help()
        print("\nAvailable facilities:", ', '.join(cli_manager.get_available_facilities()))
        sys.exit(1)

    # Check for local JSON first before any network operations
    if args.command == "make-mmcif":
        if not args.json:
            local_json = cli_manager.find_local_json(args.facility)
            if local_json:
                args.json = local_json
                print(f"Using local JSON file: {local_json}")

    if args.command == "fetch-facility-json":
        json_name = args.json_name.split('.')[0]
        available_jsons = []
        for facility, jsons in cli_manager.fetch_facilities_data().items():
            available_jsons.extend(jsons)

        available_jsons_pruned = [j.split('/')[-1].split('.')[0] for j in available_jsons] 
        if json_name not in available_jsons_pruned:
            print(f"No JSON found matching '{json_name}'")
            print("\nAvailable JSONs:")
            for json_path in available_jsons:
                print(f"  - {os.path.basename(json_path)}")
            sys.exit(1)

        index_of_match = available_jsons_pruned.index(json_name)
        cli_manager.fetch_facility_json(available_jsons[index_of_match], args.output_dir)

    elif args.command == "make-mmcif":
        available_facilities = cli_manager.get_available_facilities()
        
        if args.facility not in available_facilities:
            print(f"Invalid facility: {args.facility}")
            print(f"Available facilities: {', '.join(available_facilities)}")
            sys.exit(1)

        if not args.json:
            print("No JSON file specified and none found in current directory")
            print(f"\nAvailable JSONs for {args.facility}:")
            for json_path in cli_manager.get_facility_jsons(args.facility):
                print(f"  - {os.path.basename(json_path)}")
            print("\nFetch one using:")
            for json_path in cli_manager.get_facility_jsons(args.facility):
                base_name = os.path.splitext(os.path.basename(json_path))[0]
                print(f"  mmcif-gen fetch-facility-json {base_name}")
            sys.exit(1)

        # Create output directory
        os.makedirs(args.output_folder, exist_ok=True)
        print(f"Processing {args.facility} data using {args.json}...")

        if args.output_folder:
            pathlib.Path(args.output_folder).mkdir(parents=True, exist_ok=True) 
        if args.facility == 'pdbe':
            run_investigation_pdbe(args)
        elif args.facility == 'maxiv':
            run_investigation_maxiv(args)
        elif args.facility == 'dls':
            run_investigation_dls(args)
        elif args.facility == 'xchem':
            run_investigation_xchem(args)
        elif args.facility == 'esrf':
            run_investigation_esrf(args)

if __name__ == "__main__":
    main()


from mmcif_gen.investigation_io import InvestigationStorage
from mmcif_gen.operations import (
    operationBase,
    IntersectionOperation,
    CopyOperation,
    CopyFillOperation,
    CopyConditionalModificationOperation,
    AutoIncrementOperation,
    ConditionalUnionOperation,
    StaticValueOperation,
    ModifyOperation,
    CopyForEachRowOperation,
    NoopOperation,
    DeletionOperation,
    ExternalInformationOperation,
    ConditionalDistinctUnionOperation,
    UnionDistinctOperation,
    SQLOperation,
    EndpointOperation,
    CopyFromPickleOperation,
    JQFilterOperation
)
import json
import logging
import sys

logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)


class InvestigationEngine:
    def __init__(self, investigation_id: str, output_path: str) -> None:
        self.investigation_storage = InvestigationStorage(investigation_id)
        self.output_path = output_path
        self.investigation_id = investigation_id
        self.operations = []

    def pre_run(self) -> None:
        self.read_json_operations()


    def read_json_operations(self) -> None:
        logging.info("Reading JSON operation files")
        with open(self.operation_file_json, "r") as file:
            json_data = json.load(file)
            self.operations = json_data.get("operations", [])
            self.investigation_storage.mmcif_order = json_data.get("mmcif_order", [])

    def operation_factory(self, operation_type: str, operation_reader: str) -> operationBase:
        if not operation_reader:
            operation_reader = self.reader
        elif operation_reader == "sqlite":
            operation_reader = self.sqlite_reader
        elif operation_reader == "pickle":
            operation_reader = self.pickle_reader
        elif operation_reader == "cif":
            operation_reader = self.reader
        elif operation_reader == "json":
            operation_reader = self.json_reader


        if operation_type == "distinct_union":
            return UnionDistinctOperation(self.investigation_storage, operation_reader)
        elif operation_type == "intersection":
            return IntersectionOperation(self.investigation_storage, operation_reader)
        elif operation_type == "auto_increment":
            return AutoIncrementOperation(self.investigation_storage, operation_reader)
        elif operation_type == "static_value":
            return StaticValueOperation(self.investigation_storage, operation_reader)
        elif operation_type == "modify_intersection":
            return ModifyOperation(self.investigation_storage, operation_reader)
        elif operation_type == "conditional_union":
            return ConditionalUnionOperation(self.investigation_storage, operation_reader)
        elif operation_type == "copy":
            return CopyOperation(self.investigation_storage, operation_reader)
        elif operation_type == "copy_fill":
            return CopyFillOperation(self.investigation_storage, operation_reader)
        elif operation_type == "copy_conditional_modify":
            return CopyConditionalModificationOperation(
                self.investigation_storage, operation_reader
            )
        elif operation_type == "copy_for_each_row":
            return CopyForEachRowOperation(self.investigation_storage, operation_reader)
        elif operation_type == "external_information":
            return ExternalInformationOperation(self.investigation_storage, operation_reader)
        elif operation_type == "deletion":
            return DeletionOperation(self.investigation_storage, operation_reader)
        elif operation_type == "conditional_distinct_union":
            return ConditionalDistinctUnionOperation(
                self.investigation_storage, operation_reader
            )
        elif operation_type == "sql_query":
            return SQLOperation(self.investigation_storage, operation_reader)
        elif operation_type == "rest_endpoint":
            return EndpointOperation(self.investigation_storage, None, operation_reader)
        elif operation_type == "noop":
            return NoopOperation(self.investigation_storage, operation_reader)
        elif operation_type == "copy_from_pickle":
            return CopyFromPickleOperation(self.investigation_storage, operation_reader)
        elif operation_type == "jq_filter":
            return JQFilterOperation(self.investigation_storage, operation_reader)
        else:
            raise ValueError(f"Invalid operation type: {operation_type}")

    def run(self, prefer_pairs: bool = False) -> None :
        for operation_data in self.operations:
            try:
                operation_type = operation_data["operation"]
                operation_reader = operation_data.get("reader", None)
                operation = self.operation_factory(operation_type, operation_reader)
                operation.perform_operation(operation_data)
            except Exception as e:
                logging.error(f"Operation Failed:")
                logging.exception(e)

        self.investigation_storage.write_data_to_cif(
            f"{self.output_path}/{self.investigation_id}.cif",
            prefer_pairs=False
        )



from mmcif_gen.investigation_engine import InvestigationEngine
from mmcif_gen.investigation_io import SqliteReader,CIFReader,PickleReader
from typing import List
import sys
import os
import logging

logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)

class InvestigationXChem(InvestigationEngine):
        
    def __init__(self, sqlite_path: str, deposit_path: str, cif_files: str, investigation_id: str, output_path: str) -> None:
        logging.info("Instantiating XChem Investigation subclass")
        self.sqlite_reader= SqliteReader(sqlite_path)
        self.pickle_reader = PickleReader(deposit_path).data
        self.reader = CIFReader()
        self.reader.read_files(cif_files)
        self.operation_file_json = "./operations/xchem_operations.json"
        self.excluded_libraries = ["'Diffraction Test'","'Solvent'"]
        super().__init__(investigation_id, output_path)

    def pre_run(self) -> None:
        logging.info("Pre-running")
        libraries=["XChem_Libraries_2024-02-01.csv"]
        for library in libraries:
            self.load_library_csv(f"./external_data/{library}")
        self.create_experiment_table()
        self.find_missing_compound_information()
        super().pre_run()

    def find_missing_compound_information(self) -> None:
        missing_compounds = self.sqlite_reader.sql_execute(f"SELECT DISTINCT b.CompoundCode, b.ID, b.LibraryName FROM mainTable b LEFT JOIN `XChem_Libraries_2024-02-01` a ON b.CompoundCode = a.vendor_catalog_ID WHERE a.vendor_catalog_ID IS NULL AND b.LibraryName NOT IN ({','.join(self.excluded_libraries)})")
        logging.warning(f"Number of missing compounds: {len(missing_compounds)}")
        for compound in missing_compounds:
            logging.warning(f"Compound Code: {compound[0]}, ID: {compound[1]}, Library Name: {compound[2]}")

    def create_experiment_table(self) -> None:
        # Retrieve distinct series based on the provided query
        distinct_series = self.sqlite_reader.sql_execute('''
            SELECT DISTINCT LibraryName 
            FROM mainTable 
            WHERE CompoundSMILES IS NOT NULL 
                AND CompoundSMILES != '' 
                AND CompoundSMILES != '-'
        ''')
        
        # Create a mapping from LibraryName to series_id
        series_mapping = {}
        alloted_id = 1
        for row in distinct_series:
            if row[0] in series_mapping:
                continue
            else:
                series_mapping[row[0]] = alloted_id
                alloted_id += 1
            
        logging.info(f"Series Mapping: {series_mapping}")
        
        experimental_data = self.sqlite_reader.sql_execute(f'''
            SELECT a.CompoundCode, a.LibraryName, a.CompoundSMILES, a.RefinementOutcome, b.`Parent InChI Key` 
            FROM mainTable a 
            INNER JOIN `XChem_Libraries_2024-02-01` b 
                ON a.CompoundCode = b.vendor_catalog_ID
            WHERE a.LibraryName NOT IN ({','.join(self.excluded_libraries)});
        ''')

        # inchi_keys_mapping = {inchi_key[4]: idx + 1 for idx, inchi_key in enumerate(experimental_data)}
        inchi_keys_mapping = {}
        alloted_id = 1
        for inchi_key in experimental_data:
            if inchi_key[4] in inchi_keys_mapping:
                continue
            else:
                inchi_keys_mapping[inchi_key[4]] = alloted_id
                alloted_id += 1

        # Drop the experiments table if it exists
        self.sqlite_reader.sql_execute("DROP TABLE IF EXISTS experiments")
        
        # Create the experiments table with series_id and series columns
        self.sqlite_reader.sql_execute('''
            CREATE TABLE experiments (
                screening_exp_id INTEGER PRIMARY KEY AUTOINCREMENT, 
                investigation_id TEXT,
                sample_id INTEGER,
                campaign_id TEXT,
                series_id INTEGER,
                series TEXT,
                library_name TEXT,
                compound_smiles TEXT,
                compound_code TEXT,
                fragment_component_mix_id INTEGER,
                result_id INTEGER,
                fraglib_component_id INTEGER,
                refinement_outcome TEXT,
                outcome TEXT,
                outcome_assessment TEXT,
                outcome_description TEXT,
                outcome_details TEXT,
                data_deposited TEXT
            )
        ''')
        
        with self.sqlite_reader.sqlite_db_connection() as cursor:
            for index, experiment in enumerate(experimental_data):
                library_name = experiment[1]
                inchi_key = experiment[4]
                refinement_outcome = experiment[3]

                
                # Determine outcome based on refinement_outcome
                outcome = None
                if refinement_outcome == '5 - Deposition ready':
                    outcome = 'hit'
                elif refinement_outcome == '7 - Analysed & Rejected':
                    outcome = 'miss'
                elif refinement_outcome == '4 - CompChem ready':
                    outcome = 'partial hit'
                
                # Determine outcome_assessment based on refinement_outcome
                if refinement_outcome == '5 - Deposition ready':
                    outcome_assessment = 'manual'
                elif refinement_outcome == '7 - Analysed & Rejected':
                    outcome_assessment = 'refined'
                else:
                    outcome_assessment = 'automatic'
                
                # Determine data_deposited based on refinement_outcome
                if refinement_outcome == '6 - Deposited':
                    data_deposited = 'Y'
                else:
                    data_deposited = 'N'
                
                # Set outcome_details only if 'Analysed & Rejected'
                outcome_details = 'Analysed & Rejected' if refinement_outcome == '7 - Analysed & Rejected' else None
                
                # Retrieve series_id from the mapping
                series_id = series_mapping.get(library_name)
                series = library_name  # Assuming series is equivalent to LibraryName
                fraglib_component_id = inchi_keys_mapping.get(inchi_key)
                
                if series_id is None:
                    logging.warning(f"LibraryName '{library_name}' not found in series mapping.")
                    continue  # Skip insertion if series_id is not found
                
                # Single insertion with all fields, including series_id and series
                cursor.execute('''
                    INSERT INTO experiments (
                        investigation_id, library_name, compound_smiles, compound_code, refinement_outcome,
                        campaign_id, sample_id, fraglib_component_id,
                        outcome, outcome_assessment, outcome_details,
                        series_id, series, data_deposited
                    ) 
                    VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)
                ''', (
                    self.investigation_id, 
                    library_name, 
                    experiment[2], 
                    experiment[0], 
                    refinement_outcome,
                    1,  # campaign_id
                    1,  # sample_id
                    fraglib_component_id,
                    outcome, 
                    outcome_assessment, 
                    outcome_details,
                    series_id,
                    series,
                    data_deposited
                ))
    
    def load_library_csv(self, csv_file: str) -> None:
        # Load csv file and put this on an sqlite table in the existing sqlite_reader
        self.sqlite_reader.create_table_from_csv(csv_file)

def get_cif_file_paths(folder_path : str) -> List[str]:
    cif_file_paths = []
    for root, dirs, files in os.walk(folder_path):
        for file in files:
            if ".txt" in file:
                cif_file_paths.append(os.path.join(root, file))
    if not cif_file_paths:
        logging.warn(f"No cif files in the folder path: {folder_path}")
        raise Exception("Model file path is empty")
    return cif_file_paths


def xchem_subparser(subparsers, parent_parser):
    parser_xchem = subparsers.add_parser("xchem", help="Parameter requirements for creating investigation files from XChem data", parents=[parent_parser])

    parser_xchem.add_argument(
        "--sqlite",
        help="Path to the .sqlite file for each data set"
    )
    parser_xchem.add_argument(
        "--deposit",
        help="Path for the deposition process via XCE"
    )
    parser_xchem.add_argument(
        "--txt",
        help="Path to add additional information or overwrite in mmcifs"
    )

def run(sqlite_path : str, deposit_path: str, txt_path: str, investigation_id: str, output_path: str) -> None:
    cif_files = []
    if txt_path:
        cif_files = get_cif_file_paths(txt_path)
        print("List of CIF file paths:")
        for file_path in cif_files:
            print(file_path)
    im = InvestigationXChem(sqlite_path, deposit_path, cif_files, investigation_id, output_path)
    im.pre_run()
    im.run()

def run_investigation_xchem(args):
    if not args.sqlite:
        logging.error("XChem facility requires path to --sqlite file")
        return 1
    run(args.sqlite, args.deposit, args.txt, args.id, args.output_folder)



from mmcif_gen.investigation_engine import InvestigationEngine
from mmcif_gen.investigation_io import RestReader
from typing import List
import sys
import logging

logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)

class InvestigationESRF(InvestigationEngine):
        
    def __init__(self, esrf_json_path: str, investigation_id: str, output_path: str) -> None:
        logging.info("Instantiating ESRF Investigation subclass")
        # self.rest_reader = RestReader("https://htxlab.embl.org/", user, pw)
        # self.rest_reader.get_auth_token()
        self.operation_file_json = "./operations/esrf/esrf_investigation.json"
        self.json_reader = JsonReader(esrf_json_path)
        super().__init__(investigation_id, output_path)

    def pre_run(self) -> None:
        logging.info("Pre-running")
        super().pre_run()

def run(esrf_json_path : str, id: str, output_path: str, operation_json_path: str) -> None:
    im = InvestigationESRF(esrf_json_path, id, output_path, operation_json_path)
    im.pre_run()
    im.run() 

def run_investigation_esrf(args):
    if not args.esrf_json:
        logging.error("ESRF facility requires path to --esrf-json file generated from CRIMMS")
        return 1
    run(args.esrf_json, args.id, args.output_folder,args.json)   

def esrf_subparser(subparsers, parent_parser):
    parser_pdbe = subparsers.add_parser("esrf",help="Parameter requirements for investigation files from ESRF data", parents=[parent_parser])
    parser_pdbe.add_argument(
        "-f", 
        "--esrf-json", help="Path to the json file created from CRIMMS microservice"
    )




from mmcif_gen.investigation_engine import InvestigationEngine
from mmcif_gen.investigation_io import JsonReader
from typing import List
import sys
import logging

logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)


class InvestigationDLS(InvestigationEngine):
        
    def __init__(self, json_path: str, id: str, output_path: str, transformation_json: str="./operations/dls/dls_metadata.json") -> None:
        logging.info("Instantiating DLS subclass")
        logging.info(f"Creating file id: {id}")
        self.json_reader = JsonReader(json_path)
        self.operation_file_json = transformation_json
        super().__init__(id, output_path)

    def pre_run(self) -> None:
        logging.info("Pre-running")
        super().pre_run()


def dls_subparser(subparsers, parent_parser):
    parser_dls = subparsers.add_parser("dls", help="Parameter requirements for creating mmcif files from DLS data", parents=[parent_parser])

    parser_dls.add_argument(
        "--dls-json",
        help="Path to the .json file created from ISYPB"
    )

def run(dls_json_path : str, id: str, output_path: str, operation_json_path: str) -> None:
    im = InvestigationDLS(dls_json_path, id, output_path, operation_json_path)
    im.pre_run()
    im.run(prefer_pairs=True)

def run_investigation_dls(args):
    if not args.dls_json:
        logging.error("DLS facility requires path to --dls-json file generated from ISYPB")
        return 1
    run(args.dls_json, args.id, args.output_folder,args.json)



from mmcif_gen.investigation_engine import InvestigationEngine
from mmcif_gen.investigation_io import CIFReader, SqliteReader
import os
import requests
from typing import List, Dict
import sys
import logging
import gzip
import tempfile
import shutil
import csv
from contextlib import contextmanager
import sqlite3

logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)

FTP_URL_UPDATED = (
    "https://ftp.ebi.ac.uk/pub/databases/msd/updated_mmcif/divided/{}/{}_updated.cif.gz"
)
FTP_URL_ARCHIVE = (
    "https://ftp.ebi.ac.uk/pub/databases/pdb/data/structures/divided/mmCIF/{}/{}.cif.gz"
)

class InvestigationPdbe(InvestigationEngine):
        
    def __init__(self, model_file_path: List[str], investigation_id: str, output_path: str, pdbe_investigation_json: str="./operations/pdbe/pdbe_investigation.json") -> None:
        logging.info("Instantiating PDBe Investigation subclass")
        self.reader = CIFReader()
        self.model_file_path = model_file_path
        self.operation_file_json = "./operations/pdbe/pdbe_investigation.json"
        self.sqlite_reader = SqliteReader("pdbe_sqlite.db")
        super().__init__(investigation_id, output_path)

    def pre_run(self) -> None:
        logging.info("Pre-running")
        self.reader.read_files(self.model_file_path)
        self.create_denormalised_tables()
        self.build_denormalised_data()
        self.add_struct_ref_data()
        self.add_descript_categories()
        self.add_sample_category()
        self.add_synchrotron_data()
        self.add_exptl_data()
        self.add_investigation_id(self.investigation_id)
        super().pre_run()


    def sql_execute(self, query):
        logging.debug(f"Executing query: {query}")
        result = []
        with self.sqlite_reader.sqlite_db_connection() as conn:
            response = conn.execute(query)
            for row in response:
                result.append(row)
        return result

    def create_denormalised_tables(self):
        logging.info("Creating denormalised table")
        drop_denormalized_table = "DROP TABLE IF EXISTS denormalized_data;"
        create_denormalized_table = """
            CREATE TABLE denormalized_data (
                investigation_entity_id INT,
                pdb_id TEXT,
                model_file_no TEXT,
                file_name TEXT,
                entity_id TEXT,
                type TEXT,
                seq_one_letter_code TEXT,
                chem_comp_id TEXT,
                src_method TEXT,
                description TEXT,
                poly_type TEXT,
                poly_descript INT,
                nonpoly_descript INT,
                sample_id INT,
                db_name TEXT,
                db_code TEXT,
                db_accession TEXT,
                synchrotron_site TEXT,
                exptl_method TEXT,
                campaign_id TEXT,
                series_id TEXT,
                investigation_id TEXT
            )
        """
        with self.sqlite_reader.sqlite_db_connection() as cursor:
            cursor.execute(drop_denormalized_table)
            cursor.execute(create_denormalized_table)

    def build_denormalised_data(self):
        logging.info("Building Denormalized data table from the cif files")
        denormalized_data = []
        ordinals = {}
        next_poly_ordinal = 1
        next_nonpoly_ordinal = 1
        for file_name, datablock in self.reader.data.items():
            entity_category = datablock.find_mmcif_category("_entity")
            entity_poly_category = datablock.find_mmcif_category("_entity_poly")
            entity_nonpoly_category = datablock.find_mmcif_category(
                "_pdbx_entity_nonpoly"
            )
            database_2_category = datablock.find_mmcif_category("_database_2")

            # Create dictionaries to map column names to their indices
            entity_columns = {name: i for i, name in enumerate(entity_category.tags)}
            poly_columns = {name: i for i, name in enumerate(entity_poly_category.tags)}
            nonpoly_columns = {
                name: i for i, name in enumerate(entity_nonpoly_category.tags)
            }
            database_2_columns = {
                name: i for i, name in enumerate(database_2_category.tags)
            }

            pdb_id = database_2_category[0][
                database_2_columns["_database_2.database_code"]
            ]
            if entity_category is not None:
                for row in entity_category:
                    entity_id = row[entity_columns["_entity.id"]]
                    entity_type = row[entity_columns["_entity.type"]]
                    src_method = row[entity_columns["_entity.src_method"]]
                    description = row[entity_columns["_entity.pdbx_description"]].strip("'").strip(";").strip("\n")
                    chem_comp_id = ""
                    seq_one_letter_code = ""
                    ordinal = ""

                    if entity_type == "polymer":
                        seq_one_letter_code = ""
                        poly_type = ""
                        # Check if the entity has polymer data
                        if entity_poly_category is not None:
                            for poly_row in entity_poly_category:
                                if (
                                    poly_row[poly_columns["_entity_poly.entity_id"]]
                                    == entity_id
                                ):
                                    seq_one_letter_code = poly_row[
                                        poly_columns[
                                            "_entity_poly.pdbx_seq_one_letter_code"
                                        ]
                                    ]
                                    poly_type = poly_row[
                                        poly_columns["_entity_poly.type"]
                                    ]

                        ordinal = ordinals.get(seq_one_letter_code, False)
                        if not ordinal:
                            ordinal = next_poly_ordinal
                            ordinals[seq_one_letter_code] = next_poly_ordinal
                            next_poly_ordinal = next_poly_ordinal + 1

                    elif entity_type in ["water", "non-polymer"]:
                        # Check if the entity has non-polymer data
                        if entity_nonpoly_category is not None:
                            for nonpoly_row in entity_nonpoly_category:
                                if (
                                    nonpoly_row[
                                        nonpoly_columns[
                                            "_pdbx_entity_nonpoly.entity_id"
                                        ]
                                    ]
                                    == entity_id
                                ):
                                    chem_comp_id = nonpoly_row[
                                        nonpoly_columns["_pdbx_entity_nonpoly.comp_id"]
                                    ]
                        ordinal = ordinals.get(chem_comp_id, False)
                        if not ordinal:
                            ordinal = next_nonpoly_ordinal
                            ordinals[chem_comp_id] = ordinal
                            next_nonpoly_ordinal = next_nonpoly_ordinal + 1

                    denormalized_data.append(
                        {
                            "ordinal": ordinal,
                            "pdb_id": pdb_id,
                            "file_name": file_name,
                            "model_file_no": "",  
                            "entity_id": entity_id,
                            "type": entity_type,
                            "seq_one_letter_code": seq_one_letter_code.strip(";").rstrip('\n'),  # Placeholder for polymer data
                            "chem_comp_id": chem_comp_id,
                            "src_method": src_method,
                            "poly_type": poly_type.strip("'"),
                            "description": description,
                        }
                    )
        logging.info("Successfully built the data for the table")
        logging.info("Loading table into In-memory Sqlite")

        with self.sqlite_reader.sqlite_db_connection() as cursor:
            for row in denormalized_data:
                insert_query = """
                    INSERT INTO denormalized_data
                    (investigation_entity_id, pdb_id, file_name, model_file_no, entity_id, type, seq_one_letter_code, chem_comp_id, src_method, description, poly_type)
                    VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)
                """
                cursor.execute(
                    insert_query,
                    (
                        row["ordinal"],
                        row["pdb_id"],
                        row["file_name"],
                        row["model_file_no"],
                        row["entity_id"],
                        row["type"],
                        row["seq_one_letter_code"],
                        row["chem_comp_id"],
                        row["src_method"],
                        row["description"],
                        row["poly_type"],
                    ),
                )

    def add_descript_categories(self):
        logging.info("Adding descript categories info to the table")
        poly_descript = {}
        non_poly_descript = {}

        unique_poly = self.sql_execute(
            """
            SELECT DISTINCT(set_of_poly) FROM
                (
                    SELECT pdb_id,
                    GROUP_CONCAT(investigation_entity_id) AS set_of_poly
                    FROM 
                    (
                        SELECT pdb_id, investigation_entity_id FROM denormalized_data WHERE type="polymer" ORDER BY investigation_entity_id
                    )
                    GROUP BY pdb_id
                )
                GROUP BY set_of_poly
            """
        )

        for i, poly in enumerate(unique_poly):
            poly_descript[poly[0]] = i + 1

        all_poly_groups = self.sql_execute(
            """
            SELECT pdb_id,
            GROUP_CONCAT(investigation_entity_id) AS set_of_poly
            FROM 
            (
                SELECT pdb_id, investigation_entity_id FROM denormalized_data WHERE type="polymer" ORDER BY investigation_entity_id
            )
            GROUP BY pdb_id
            """
        )

        for group in all_poly_groups:
            pdb_id = group[0]
            poly_descript_id = poly_descript[group[1]]
            self.sql_execute(
                f"""
                UPDATE denormalized_data
                SET poly_descript = {poly_descript_id}
                WHERE pdb_id = "{pdb_id}"
                """
            )

        unique_non_poly = self.sql_execute(
            """
                SELECT DISTINCT(set_of_non_poly) FROM
                    (
                        SELECT pdb_id,
                        GROUP_CONCAT(investigation_entity_id) AS set_of_non_poly
                        FROM 
                        (
                            SELECT pdb_id, investigation_entity_id FROM denormalized_data WHERE type="non-polymer" OR type="water" ORDER BY investigation_entity_id
                        )
                        GROUP BY pdb_id
                    )
                    GROUP BY set_of_non_poly
                """
        )

        for i, non_poly in enumerate(unique_non_poly):
            non_poly_descript[non_poly[0]] = i + 1

        all_nonpoly_groups = self.sql_execute(
            """
            SELECT pdb_id,
            GROUP_CONCAT(investigation_entity_id) AS set_of_non_poly
            FROM 
            (
                SELECT pdb_id, investigation_entity_id FROM denormalized_data WHERE type="non-polymer" OR type="water" ORDER BY investigation_entity_id
            )
            GROUP BY pdb_id
            """
        )

        for group in all_nonpoly_groups:
            pdb_id = group[0]
            non_poly_descript_id = non_poly_descript[group[1]]
            self.sql_execute(
                f"""
                            UPDATE denormalized_data
                            SET nonpoly_descript = {non_poly_descript_id}
                            WHERE pdb_id = "{pdb_id}"
                            """
            )

    def add_sample_category(self):
        unique_samples = self.sql_execute(
            """
                        SELECT poly_descript, nonpoly_descript from 
                        denormalized_data GROUP BY poly_descript, nonpoly_descript"""
        )
        for index, sample in enumerate(unique_samples):
            self.sql_execute(
                f"""
                            UPDATE denormalized_data
                            SET sample_id = {index+1}
                            WHERE poly_descript = "{sample[0]}" AND 
                            nonpoly_descript = "{sample[1]}"
                            """
            )

    def add_exptl_data(self):
        exptl_data = []
        for file_name, datablock in self.reader.data.items():
            exptl_category = datablock.find_mmcif_category("_exptl")
            exptl_columns = {name: i for i, name in enumerate(exptl_category.tags)}
            database_2_category = datablock.find_mmcif_category("_database_2")
            database_2_columns = {
                name: i for i, name in enumerate(database_2_category.tags)
            }
            pdb_id = database_2_category[0][
                database_2_columns["_database_2.database_code"]
            ]

            if exptl_category is not None:
                for row in exptl_category:
                    exptl_data.append(
                        {
                            "pdb_id": pdb_id,
                            "exptl_method": row[exptl_columns["_exptl.method"]].strip("'"),
                        }
                    )
        for row in exptl_data:
            self.sql_execute(
                f"""
                            UPDATE denormalized_data
                            SET 
                                exptl_method = {repr(row['exptl_method'])}
                            WHERE 
                                pdb_id = "{row['pdb_id']}" 
                            """
            )

    def add_synchrotron_data(self):
        synchrotron_data = []
        campaigns = {}
        next_ordinal = 1
        for file_name, datablock in self.reader.data.items():
            diffrn_source_category = datablock.find_mmcif_category("_diffrn_source")
            diffrn_source_columns = {
                name: i for i, name in enumerate(diffrn_source_category.tags)
            }
            database_2_category = datablock.find_mmcif_category("_database_2")
            database_2_columns = {
                name: i for i, name in enumerate(database_2_category.tags)
            }
            pdb_id = database_2_category[0][
                database_2_columns["_database_2.database_code"]
            ]
            if diffrn_source_category is not None:
                for row in diffrn_source_category:
                    synchrotron_site = row[
                        diffrn_source_columns["_diffrn_source.pdbx_synchrotron_site"]
                    ]
                    if synchrotron_site not in campaigns:
                        campaigns[synchrotron_site] = next_ordinal
                        next_ordinal = next_ordinal + 1
                    synchrotron_data.append(
                        {
                            "pdb_id": pdb_id,
                            "synchrotron_site": synchrotron_site,
                            "campaign_id": campaigns[synchrotron_site],
                            "series_id": campaigns[synchrotron_site],
                        }
                    )
        for row in synchrotron_data:
            self.sql_execute(
                f"""
                            UPDATE denormalized_data
                            SET 
                                synchrotron_site = "{row['synchrotron_site']}",
                                campaign_id = {row['campaign_id']},
                                series_id = {row['series_id']}
                            WHERE 
                                pdb_id = "{row['pdb_id']}" 
                            """
            )

    def add_struct_ref_data(self):
        struct_ref = []
        for file_name, datablock in self.reader.data.items():
            struct_ref_category = datablock.find_mmcif_category("_struct_ref")
            database_2_category = datablock.find_mmcif_category("_database_2")
            struct_ref_columns = {
                name: i for i, name in enumerate(struct_ref_category.tags)
            }
            database_2_columns = {
                name: i for i, name in enumerate(database_2_category.tags)
            }
            pdb_id = database_2_category[0][
                database_2_columns["_database_2.database_code"]
            ]

            if struct_ref_category is not None:
                for row in struct_ref_category:
                    struct_ref.append(
                        {
                            "pdb_id": pdb_id,
                            "entity_id": row[
                                struct_ref_columns["_struct_ref.entity_id"]
                            ],
                            "db_name": row[struct_ref_columns["_struct_ref.db_name"]],
                            "db_code": row[struct_ref_columns["_struct_ref.db_code"]],
                            "pdbx_db_accession": row[
                                struct_ref_columns["_struct_ref.pdbx_db_accession"]
                            ],
                        }
                    )

        for row in struct_ref:
            self.sql_execute(
                f"""
                            UPDATE denormalized_data
                            SET 
                             db_name = "{row['db_name']}",
                             db_code = "{row['db_code']}",
                             db_accession = "{row['pdbx_db_accession']}"
                            WHERE 
                            pdb_id = "{row['pdb_id']}" AND 
                            entity_id = "{row['entity_id']}" AND
                            type = "polymer"
                            """
            )

    def add_investigation_id(self, investigation_id: str):
        self.sql_execute(
            f"""
                            UPDATE denormalized_data
                            SET investigation_id = "{investigation_id}"
                            """
        )



def download_and_run_pdbe_investigation(pdb_ids: List[str], investigation_id: str, output_path:str, json_path: str) -> None:
    logging.info(f"Creating investigation files for pdb ids: {pdb_ids}")
    temp_dir = tempfile.mkdtemp()
    try:
        for pdb_code in pdb_ids:
            url = FTP_URL_ARCHIVE.format(pdb_code[1:3], pdb_code)

            compressed_file_path = os.path.join(temp_dir, f"{pdb_code}.cif.gz")
            uncompressed_file_path = os.path.join(temp_dir, f"{pdb_code}.cif")

            response = requests.get(url)
            if response.status_code == 200:
                with open(compressed_file_path, "wb") as f:
                    f.write(response.content)

                with gzip.open(compressed_file_path, "rb") as gz:
                    with open(uncompressed_file_path, "wb") as f:
                        f.write(gz.read())
                logging.info(f"Downloaded and unzipped {pdb_code}.cif")
            else:
                logging.info(f"Failed to download {pdb_code}.cif.gz")

        run(temp_dir, investigation_id, output_path, json_path)

    except Exception as e:
        logging.exception(f"An error occurred: {str(e)}")

    finally:
        for pdb_code in pdb_ids:
            compressed_file_path = os.path.join(temp_dir, f"{pdb_code}.cif.gz")
            uncompressed_file_path = os.path.join(temp_dir, f"{pdb_code}.cif")
            if os.path.exists(compressed_file_path):
                os.remove(compressed_file_path)
            if os.path.exists(uncompressed_file_path):
                os.remove(uncompressed_file_path)

        shutil.rmtree(temp_dir)

def run_investigation_pdbe(args):
    if args.model_folder:
        run(args.model_folder, args.id,args.output_folder, args.json)
    elif args.pdb_ids:
        download_and_run_pdbe_investigation(args.pdb_ids, args.investigation_id, args.output_folder, args.json)
    elif args.csv_file:
        group_data = parse_csv(args.csv_file)
        for group, entry in group_data.items():
            try:
                download_and_run_pdbe_investigation(entry, group, args.output_folder, args.json)
            except Exception as e:
                logging.exception(e)
    else:
        logging.error("PDBe Facilitiy requires parameter: --model-folder OR --csv-file OR --pdb-ids ")


def get_cif_file_paths(folder_path : str) -> List[str]:
    cif_file_paths = []
    for root, dirs, files in os.walk(folder_path):
        for file in files:
            if ".cif" in file and ".gz" not in file:
                cif_file_paths.append(os.path.join(root, file))
    if not cif_file_paths:
        logging.warn(f"No cif files in the folder path: {folder_path}")
        raise Exception("Model file path is empty")
    return cif_file_paths


def run(folder_path : str, investigation_id: str, output_path: str, json_path: str) -> None:
    model_file_path = get_cif_file_paths(folder_path)
    print("List of CIF file paths:")
    for file_path in model_file_path:
        print(file_path)
    im = InvestigationPdbe(model_file_path, investigation_id, output_path, json_path)
    im.pre_run()
    im.run()

def parse_csv(csv_file:str) -> Dict:
    group_data = {}
    with open(csv_file) as file:
        csv_reader = csv.DictReader(file, delimiter=",")
        for row in csv_reader:
            group_id = row["GROUP_ID"]
            entry_id = row["ENTRY_ID"]

            if group_id in group_data:
                group_data[group_id].append(entry_id)
            else:
                group_data[group_id] = [entry_id]
    return group_data
    
def pdbe_subparser(subparsers, parent_parser):
    parser_pdbe = subparsers.add_parser("pdbe",help="Parameter requirements for investigation files from PDBe data", parents=[parent_parser])
    parser_pdbe.add_argument(
        "-f", 
        "--model-folder", help="Directory which contains model files"
    )
    parser_pdbe.add_argument(
        "-csv", 
        "--csv-file", help="Requires CSV with 2 columns [GROUP_ID, ENTRY_ID]"
    )
    parser_pdbe.add_argument(
        "-p",
        "--pdb-ids",
        nargs="+",
        help="Create investigation from set of pdb ids, space seperated",
    )


import argparse
import gemmi
import sys
import logging
from mmcif_gen.util.output_grabber import OutputGrabber
import tempfile
import os
import requests
from typing import Dict, List
import gzip
import shutil
import csv
import pathlib

try:
    from openbabel import pybel
except ImportError:
    import pybel  

FTP_URL_ARCHIVE_SF = (
    "https://ftp.ebi.ac.uk/pub/databases/pdb/data/structures/divided/structure_factors/{}/r{}sf.ent.gz"
)


logging.basicConfig(stream=sys.stdout, level=logging.INFO)
    

def sanitize_gemmi_output(row):
    for index, item in enumerate(row):
        if isinstance(item,str):
            row[index] = gemmi.cif.as_string(item)
            if item == '?':
                row[index] = None
    return row

def smiles_to_inchikey_openbabel(smiles):
    out_stderr = OutputGrabber(sys.stderr)
    with out_stderr:
        try:
            mol = pybel.readstring("smi", smiles)
            inchikey = mol.write("inchikey").strip()
            logging.debug(f"SMILE: {smiles}")
            logging.debug(f"INCHI: {inchikey}")
            logging.debug(f"=================================")
            return inchikey
        except Exception as e:
            logging.error("Conversion failure")

def temp_download_and_process_file(investigation_cif: str, pdb_code :str ) -> None:
    logging.info(f"Creating investigation files for pdb ids: {pdb_code}")
    temp_dir = tempfile.mkdtemp()
    try:
        url = FTP_URL_ARCHIVE_SF.format(pdb_code[1:3], pdb_code)

        compressed_file_path = os.path.join(temp_dir, f"r{pdb_code}sf.ent.gz")
        uncompressed_file_path = os.path.join(temp_dir, f"r{pdb_code}sf.ent")

        response = requests.get(url)
        if response.status_code == 200:
            with open(compressed_file_path, "wb") as f:
                f.write(response.content)

            with gzip.open(compressed_file_path, "rb") as gz:
                with open(uncompressed_file_path, "wb") as f:
                    f.write(gz.read())
            logging.info(f"Downloaded and unzipped r{pdb_code}sf.ent")
        else:
            logging.info(f"Failed to download r{pdb_code}sf.ent.gz")

        process_mmcif_files(investigation_cif, uncompressed_file_path)

    except Exception as e:
        logging.exception(f"An error occurred: {str(e)}")

    finally:
        compressed_file_path = os.path.join(temp_dir, f"{pdb_code}.cif.gz")
        uncompressed_file_path = os.path.join(temp_dir, f"{pdb_code}.cif")
        if os.path.exists(compressed_file_path):
            os.remove(compressed_file_path)
        if os.path.exists(uncompressed_file_path):
            os.remove(uncompressed_file_path)

        shutil.rmtree(temp_dir)

def download_and_process_file(investigation_cif: str, pdb_code :str ) -> None:
    logging.info(f"Creating investigation files for pdb ids: {pdb_code}")
    dir = "./sf-files"
    pathlib.Path(dir).mkdir(parents=True, exist_ok=True) 
    uncompressed_file_path = pathlib.Path(f"{dir}/r{pdb_code}sf.ent")
    already_downloaded = pathlib.Path(uncompressed_file_path).is_file()
    try:
        if not already_downloaded:
            logging.info("File does not exist. Downloading...")
            url = FTP_URL_ARCHIVE_SF.format(pdb_code[1:3], pdb_code)

            compressed_file_path = os.path.join(dir, f"r{pdb_code}sf.ent.gz")
            uncompressed_file_path = os.path.join(dir, f"r{pdb_code}sf.ent")

            response = requests.get(url)
            if response.status_code == 200:
                with open(compressed_file_path, "wb") as f:
                    f.write(response.content)

                with gzip.open(compressed_file_path, "rb") as gz:
                    with open(uncompressed_file_path, "wb") as f:
                        f.write(gz.read())
                logging.info(f"Downloaded and unzipped r{pdb_code}sf.ent")
            else:
                logging.info(f"Failed to download r{pdb_code}sf.ent.gz")

        process_mmcif_files(investigation_cif, str(uncompressed_file_path))

    except Exception as e:
        logging.exception(f"An error occurred: {str(e)}")

def process_mmcif_files(investigation_cif, sf_file_cif):
    logging.info("Processing Structure factor file")
    sf_file = gemmi.cif.read(sf_file_cif)
    investigation = gemmi.cif.read(investigation_cif)

    inchi_keys = set()
    logging.info("Finding Smile string from sf file and converting them to Inchi keys")
    for block in sf_file:
        try:
            details_item = block.find_value("_diffrn.details")
            
            if details_item:
                smiles_string = details_item.split()[-1][:-1]
                inchi_keys.add(smiles_to_inchikey_openbabel(smiles_string))
        except:
            logging.exception(e)
            continue
    logging.info(f"Found and converted {len(inchi_keys)} smiles strings to Inchi key")

    logging.info("Finding existing Inchi keys in Investigation file ")
    for block_a in investigation:
        existing_inchis_in_investigation = block_a.get_mmcif_category("_pdbx_fraghub_investigation_fraglib_component")["inchi_descriptor"]
        existing_inchis = set(existing_inchis_in_investigation)

    logging.info(f"Found {len(existing_inchis)} Inchi keys in Investigation file")
    
    repeats = existing_inchis.intersection(inchi_keys)
    if repeats:
        logging.info(f"{len(repeats)} Inchi keys in sf file already exist in investigation files:")
        for inchi in repeats:
            logging.debug(inchi)
    else:
        logging.info("No Overlapping Inchi keys between sf file and investigation file")

    inchi_keys_to_add = inchi_keys - existing_inchis

    # Writing new inchis in investigation file (4 categories)
    fraglib_category = block_a.find_mmcif_category("_pdbx_fraghub_investigation_fraglib_component")
    fraglib_columns = {name: i for i, name in enumerate(fraglib_category.tags)}
    existing_fraglib_len = len(fraglib_category)

    component_mix_category = block_a.find_mmcif_category("_pdbx_fraghub_investigation_frag_component_mix")
    component_mix_columns = {name: i for i, name in enumerate(component_mix_category.tags)}
    existing_highest_id = int(max(block_a.get_mmcif_category("_pdbx_fraghub_investigation_frag_component_mix")["id"], key=int))

    existing_mix_category_len = len(component_mix_category)

    screening_exp_category = block_a.find_mmcif_category("_pdbx_fraghub_investigation_screening_exp")
    screening_exp_columns = {name: i for i, name in enumerate(screening_exp_category.tags)}
    screening_exp_category_len = len(screening_exp_category)
    screening_exp_template =  list(screening_exp_category[0])

    screening_result_category = block_a.find_mmcif_category("_pdbx_fraghub_investigation_screening_result")
    screening_result_columns = {name: i for i, name in enumerate(screening_result_category.tags)}
    screening_result_category_len = len(screening_result_category)
    screening_result_template =  list(screening_result_category[0])

    logging.info("Adding items to the Cif categories")
    for index, inchi in enumerate(inchi_keys_to_add):
        inchi_index = str(index+existing_fraglib_len+1)
        row = [None] * fraglib_category.width()
        row[fraglib_columns["_pdbx_fraghub_investigation_fraglib_component.inchi_descriptor"]] = inchi
        row[fraglib_columns["_pdbx_fraghub_investigation_fraglib_component.id"]] = inchi_index
        row = gemmi.cif.quote_list(row)
        fraglib_category.append_row(row)


        row = [None]*component_mix_category.width()
        component_mix_index = str(existing_highest_id+index+1)
        row[component_mix_columns["_pdbx_fraghub_investigation_frag_component_mix.id"]]= component_mix_index
        row[component_mix_columns["_pdbx_fraghub_investigation_frag_component_mix.fraglib_component_id"]]= inchi_index
        row = gemmi.cif.quote_list(row)
        component_mix_category.append_row([component_mix_index, inchi_index])


        row = screening_exp_template
        row = sanitize_gemmi_output(row)

        screening_exp_index= str(index+screening_exp_category_len+1)
        row[screening_exp_columns["_pdbx_fraghub_investigation_screening_exp.frag_component_mix_id"]]= component_mix_index
        row[screening_exp_columns["_pdbx_fraghub_investigation_screening_exp.screening_exp_id"]]= screening_exp_index
        row = gemmi.cif.quote_list(row)
        screening_exp_category.append_row(row)

        row = screening_result_template
        row = sanitize_gemmi_output(row)
        screening_result_index= str(index+screening_result_category_len+1)

        row[screening_result_columns["_pdbx_fraghub_investigation_screening_result.screening_exp_id"]]= screening_exp_index
        row[screening_result_columns["_pdbx_fraghub_investigation_screening_result.result_id"]]= screening_result_index
        row[screening_result_columns["_pdbx_fraghub_investigation_screening_result.outcome"]]= "miss"
        row[screening_result_columns["_pdbx_fraghub_investigation_screening_result.fraglib_component_id"]] = component_mix_index #should this be ?
        row[screening_result_columns["_pdbx_fraghub_investigation_screening_result.outcome_description"]] = "Fragment unobserved"
        row = gemmi.cif.quote_list(row)
        screening_result_category.append_row(row)

    out_file_name = investigation_cif.split("/")[-1].split(".")[0]
    file_path = investigation_cif.split(f"/{out_file_name}")[0]
    logging.info(f"Writing out file: {file_path}/{out_file_name}_m.cif")
    investigation.write_file(f"{file_path}{out_file_name}_m.cif")


def main() -> None:
    parser = argparse.ArgumentParser(
        prog="Ground state file importer ",
        description="This utility takes as an input investigation file, and sf file. And imports the data for all the misses from the sf file and adds that to the investigation file ",
    )
    parser.add_argument(
        "-inv", "--investigation-file", help="Path to investigation file"
    )
    parser.add_argument(
        "-sf", "--sf-file", help="Path to structure factor file"
    )
    parser.add_argument(
        "-p",
        "--pdb-id",
        help="PDB ID to lookup to download the sf file",
    )
    parser.add_argument(
        "-f", "--csv-file", help="Requires CSV with 2 columns [investigation_file, Pdb Code (to fetch sf file)]"
    )
    parser.add_argument(
        "-st", "--store-sf", action='store_true', help="When used, the downloaded sf files are not deleted but stored in sf-files folder"
    )
    parser.add_argument(
        "-ir", "--csv-inv-root",  help="Specifies the root path to the investigation files, that are listed in the csv"
    )

    args = parser.parse_args()
    if args.sf_file:
        process_mmcif_files(args.investigation_file, args.sf_file)
    elif args.pdb_id:
        if not os.path.exists(args.investigation_file):
            logging.error(f"{args.investigation_file} does not exist. Enter a valid path to investigation file")
            return
        download_and_process_file(args.investigation_file, args.pdb_id)
    elif args.csv_file:
        with open(args.csv_file) as file:
            csv_reader = csv.DictReader(file, delimiter=",")
            for row in csv_reader:
                investigation_file = row["INVESTIGATION_FILE"]
                sf_file = row["SF_FILE"]
                investigation_path = f"{args.csv_inv_root}/{investigation_file}"
                if not os.path.exists(investigation_path):
                    logging.error(f"{investigation_path} does not exist. Skipping the row in the csv")
                    continue
                try:
                    if len(sf_file) == 4:
                        download_and_process_file(investigation_path, sf_file)
                    else:
                        process_mmcif_files(investigation_path, sf_file)
                except Exception as e:
                    logging.exception(e)


if __name__ == "__main__":
    main()